From davem@davemloft.net Thu Jun  2 13:09:39 2005
Date: Thu, 02 Jun 2005 12:55:50 -0700 (PDT)
Message-Id: <20050602.125550.85413412.davem@davemloft.net>
From: "David S. Miller" <davem@davemloft.net>
To: gregkh@suse.de
Subject: PCI: DMA bursting advice


After seeing, at best, "guesses" as to the following kind
of information in several drivers, I decided that we really
need a way for platforms to specifically give advice in this
area for what works best with their PCI controller implementation.

Basically, this new interface gives DMA bursting advice on
PCI.  There are three forms of the advice:

1) Burst as much as possible, it is not necessary to end bursts
   on some particular boundary for best performance.

2) Burst on some byte count multiple.  A DMA burst to some multiple of
   number of bytes may be done, but it is important to end the burst
   on an exact multiple for best performance.

   The best example of this I am aware of are the PPC64 PCI
   controllers, where if you end a burst mid-cacheline then
   chip has to refetch the data and the IOMMU translations
   which hurts performance a lot.

3) Burst on a single byte count multiple.  Bursts shall end
   exactly on the next multiple boundary for best performance.

   Sparc64 and Alpha's PCI controllers operate this way.  They
   disconnect any device which tries to burst across a cacheline
   boundary.

   Actually, newer sparc64 PCI controllers do not have this behavior.
   That is why the "pdev" is passed into the interface, so I can
   add code later to check which PCI controller the system is using
   and give advice accordingly.

Signed-off-by: David S. Miller <davem@davemloft.net>
Signed-off-by: Greg Kroah-Hartman <gregkh@suse.de>

---
 include/asm-alpha/pci.h   |   17 +++++++++++++++++
 include/asm-arm/pci.h     |    8 ++++++++
 include/asm-frv/pci.h     |    8 ++++++++
 include/asm-i386/pci.h    |    8 ++++++++
 include/asm-ia64/pci.h    |   17 +++++++++++++++++
 include/asm-mips/pci.h    |    8 ++++++++
 include/asm-parisc/pci.h  |   17 +++++++++++++++++
 include/asm-ppc/pci.h     |    8 ++++++++
 include/asm-ppc64/pci.h   |   17 +++++++++++++++++
 include/asm-sh/pci.h      |    8 ++++++++
 include/asm-sh64/pci.h    |    8 ++++++++
 include/asm-sparc/pci.h   |    8 ++++++++
 include/asm-sparc64/pci.h |   17 +++++++++++++++++
 include/asm-v850/pci.h    |    8 ++++++++
 include/asm-x86_64/pci.h  |    8 ++++++++
 include/linux/pci.h       |    9 +++++++++
 16 files changed, 174 insertions(+)

--- gregkh-2.6.orig/include/asm-alpha/pci.h	2005-06-01 13:51:23.000000000 -0700
+++ gregkh-2.6/include/asm-alpha/pci.h	2005-06-05 22:47:10.000000000 -0700
@@ -223,6 +223,23 @@
 	/* Nothing to do. */
 }
 
+static inline void pci_dma_burst_advice(struct pci_dev *pdev,
+					enum pci_dma_burst_strategy *strat,
+					unsigned long *strategy_parameter)
+{
+	unsigned long cacheline_size;
+	u8 byte;
+
+	pci_read_config_byte(pdev, PCI_CACHE_LINE_SIZE, &byte);
+	if (byte == 0)
+		cacheline_size = 1024;
+	else
+		cacheline_size = (int) byte * 4;
+
+	*strat = PCI_DMA_BURST_BOUNDARY;
+	*strategy_parameter = cacheline_size;
+}
+
 /* TODO: integrate with include/asm-generic/pci.h ? */
 static inline int pci_get_legacy_ide_irq(struct pci_dev *dev, int channel)
 {
--- gregkh-2.6.orig/include/asm-arm/pci.h	2005-03-01 23:37:53.000000000 -0800
+++ gregkh-2.6/include/asm-arm/pci.h	2005-06-05 22:47:10.000000000 -0700
@@ -42,6 +42,14 @@
 #define pci_unmap_len(PTR, LEN_NAME)		((PTR)->LEN_NAME)
 #define pci_unmap_len_set(PTR, LEN_NAME, VAL)	(((PTR)->LEN_NAME) = (VAL))
 
+static inline void pci_dma_burst_advice(struct pci_dev *pdev,
+					enum pci_dma_burst_strategy *strat,
+					unsigned long *strategy_parameter)
+{
+	*strat = PCI_DMA_BURST_INFINITY;
+	*strategy_parameter = ~0UL;
+}
+
 #define HAVE_PCI_MMAP
 extern int pci_mmap_page_range(struct pci_dev *dev, struct vm_area_struct *vma,
                                enum pci_mmap_state mmap_state, int write_combine);
--- gregkh-2.6.orig/include/asm-frv/pci.h	2005-03-01 23:38:32.000000000 -0800
+++ gregkh-2.6/include/asm-frv/pci.h	2005-06-05 22:47:10.000000000 -0700
@@ -57,6 +57,14 @@
  */
 #define PCI_DMA_BUS_IS_PHYS	(1)
 
+static inline void pci_dma_burst_advice(struct pci_dev *pdev,
+					enum pci_dma_burst_strategy *strat,
+					unsigned long *strategy_parameter)
+{
+	*strat = PCI_DMA_BURST_INFINITY;
+	*strategy_parameter = ~0UL;
+}
+
 /*
  *	These are pretty much arbitary with the CoMEM implementation.
  *	We have the whole address space to ourselves.
--- gregkh-2.6.orig/include/asm-i386/pci.h	2005-03-01 23:38:07.000000000 -0800
+++ gregkh-2.6/include/asm-i386/pci.h	2005-06-05 22:47:10.000000000 -0700
@@ -99,6 +99,14 @@
 {
 }
 
+static inline void pci_dma_burst_advice(struct pci_dev *pdev,
+					enum pci_dma_burst_strategy *strat,
+					unsigned long *strategy_parameter)
+{
+	*strat = PCI_DMA_BURST_INFINITY;
+	*strategy_parameter = ~0UL;
+}
+
 #endif /* __KERNEL__ */
 
 /* implement the pci_ DMA API in terms of the generic device dma_ one */
--- gregkh-2.6.orig/include/asm-ia64/pci.h	2005-06-01 13:51:24.000000000 -0700
+++ gregkh-2.6/include/asm-ia64/pci.h	2005-06-05 22:47:10.000000000 -0700
@@ -82,6 +82,23 @@
 #define sg_dma_len(sg)		((sg)->dma_length)
 #define sg_dma_address(sg)	((sg)->dma_address)
 
+static inline void pci_dma_burst_advice(struct pci_dev *pdev,
+					enum pci_dma_burst_strategy *strat,
+					unsigned long *strategy_parameter)
+{
+	unsigned long cacheline_size;
+	u8 byte;
+
+	pci_read_config_byte(pdev, PCI_CACHE_LINE_SIZE, &byte);
+	if (byte == 0)
+		cacheline_size = 1024;
+	else
+		cacheline_size = (int) byte * 4;
+
+	*strat = PCI_DMA_BURST_MULTIPLE;
+	*strategy_parameter = cacheline_size;
+}
+
 #define HAVE_PCI_MMAP
 extern int pci_mmap_page_range (struct pci_dev *dev, struct vm_area_struct *vma,
 				enum pci_mmap_state mmap_state, int write_combine);
--- gregkh-2.6.orig/include/asm-mips/pci.h	2005-06-01 13:51:24.000000000 -0700
+++ gregkh-2.6/include/asm-mips/pci.h	2005-06-05 22:47:10.000000000 -0700
@@ -130,6 +130,14 @@
 extern void pci_dac_dma_sync_single_for_device(struct pci_dev *pdev,
 	dma64_addr_t dma_addr, size_t len, int direction);
 
+static inline void pci_dma_burst_advice(struct pci_dev *pdev,
+					enum pci_dma_burst_strategy *strat,
+					unsigned long *strategy_parameter)
+{
+	*strat = PCI_DMA_BURST_INFINITY;
+	*strategy_parameter = ~0UL;
+}
+
 extern void pcibios_resource_to_bus(struct pci_dev *dev,
 	struct pci_bus_region *region, struct resource *res);
 
--- gregkh-2.6.orig/include/asm-parisc/pci.h	2005-06-01 13:51:24.000000000 -0700
+++ gregkh-2.6/include/asm-parisc/pci.h	2005-06-05 22:47:10.000000000 -0700
@@ -230,6 +230,23 @@
 /* export the pci_ DMA API in terms of the dma_ one */
 #include <asm-generic/pci-dma-compat.h>
 
+static inline void pci_dma_burst_advice(struct pci_dev *pdev,
+					enum pci_dma_burst_strategy *strat,
+					unsigned long *strategy_parameter)
+{
+	unsigned long cacheline_size;
+	u8 byte;
+
+	pci_read_config_byte(pdev, PCI_CACHE_LINE_SIZE, &byte);
+	if (byte == 0)
+		cacheline_size = 1024;
+	else
+		cacheline_size = (int) byte * 4;
+
+	*strat = PCI_DMA_BURST_MULTIPLE;
+	*strategy_parameter = cacheline_size;
+}
+
 extern void
 pcibios_resource_to_bus(struct pci_dev *dev, struct pci_bus_region *region,
 			 struct resource *res);
--- gregkh-2.6.orig/include/asm-ppc/pci.h	2005-06-05 22:30:35.000000000 -0700
+++ gregkh-2.6/include/asm-ppc/pci.h	2005-06-05 22:47:10.000000000 -0700
@@ -69,6 +69,14 @@
 #define pci_unmap_len(PTR, LEN_NAME)		(0)
 #define pci_unmap_len_set(PTR, LEN_NAME, VAL)	do { } while (0)
 
+static inline void pci_dma_burst_advice(struct pci_dev *pdev,
+					enum pci_dma_burst_strategy *strat,
+					unsigned long *strategy_parameter)
+{
+	*strat = PCI_DMA_BURST_INFINITY;
+	*strategy_parameter = ~0UL;
+}
+
 /*
  * At present there are very few 32-bit PPC machines that can have
  * memory above the 4GB point, and we don't support that.
--- gregkh-2.6.orig/include/asm-ppc64/pci.h	2005-06-05 22:30:35.000000000 -0700
+++ gregkh-2.6/include/asm-ppc64/pci.h	2005-06-05 22:47:10.000000000 -0700
@@ -78,6 +78,23 @@
 	return 0;
 }
 
+static inline void pci_dma_burst_advice(struct pci_dev *pdev,
+					enum pci_dma_burst_strategy *strat,
+					unsigned long *strategy_parameter)
+{
+	unsigned long cacheline_size;
+	u8 byte;
+
+	pci_read_config_byte(pdev, PCI_CACHE_LINE_SIZE, &byte);
+	if (byte == 0)
+		cacheline_size = 1024;
+	else
+		cacheline_size = (int) byte * 4;
+
+	*strat = PCI_DMA_BURST_MULTIPLE;
+	*strategy_parameter = cacheline_size;
+}
+
 extern int pci_domain_nr(struct pci_bus *bus);
 
 /* Decide whether to display the domain number in /proc */
--- gregkh-2.6.orig/include/asm-sh/pci.h	2005-03-01 23:38:13.000000000 -0800
+++ gregkh-2.6/include/asm-sh/pci.h	2005-06-05 22:47:10.000000000 -0700
@@ -96,6 +96,14 @@
 #define sg_dma_address(sg)	(virt_to_bus((sg)->dma_address))
 #define sg_dma_len(sg)		((sg)->length)
 
+static inline void pci_dma_burst_advice(struct pci_dev *pdev,
+					enum pci_dma_burst_strategy *strat,
+					unsigned long *strategy_parameter)
+{
+	*strat = PCI_DMA_BURST_INFINITY;
+	*strategy_parameter = ~0UL;
+}
+
 /* Board-specific fixup routines. */
 extern void pcibios_fixup(void);
 extern void pcibios_fixup_irqs(void);
--- gregkh-2.6.orig/include/asm-sh64/pci.h	2005-03-01 23:38:33.000000000 -0800
+++ gregkh-2.6/include/asm-sh64/pci.h	2005-06-05 22:47:10.000000000 -0700
@@ -86,6 +86,14 @@
 #define sg_dma_address(sg)	((sg)->dma_address)
 #define sg_dma_len(sg)		((sg)->length)
 
+static inline void pci_dma_burst_advice(struct pci_dev *pdev,
+					enum pci_dma_burst_strategy *strat,
+					unsigned long *strategy_parameter)
+{
+	*strat = PCI_DMA_BURST_INFINITY;
+	*strategy_parameter = ~0UL;
+}
+
 /* Board-specific fixup routines. */
 extern void pcibios_fixup(void);
 extern void pcibios_fixup_irqs(void);
--- gregkh-2.6.orig/include/asm-sparc/pci.h	2005-03-01 23:38:17.000000000 -0800
+++ gregkh-2.6/include/asm-sparc/pci.h	2005-06-05 22:47:10.000000000 -0700
@@ -144,6 +144,14 @@
 
 #define pci_dac_dma_supported(dev, mask)	(0)
 
+static inline void pci_dma_burst_advice(struct pci_dev *pdev,
+					enum pci_dma_burst_strategy *strat,
+					unsigned long *strategy_parameter)
+{
+	*strat = PCI_DMA_BURST_INFINITY;
+	*strategy_parameter = ~0UL;
+}
+
 static inline void pcibios_add_platform_entries(struct pci_dev *dev)
 {
 }
--- gregkh-2.6.orig/include/asm-sparc64/pci.h	2005-06-01 13:51:25.000000000 -0700
+++ gregkh-2.6/include/asm-sparc64/pci.h	2005-06-05 22:47:10.000000000 -0700
@@ -220,6 +220,23 @@
 	return (dma_addr == PCI_DMA_ERROR_CODE);
 }
 
+static inline void pci_dma_burst_advice(struct pci_dev *pdev,
+					enum pci_dma_burst_strategy *strat,
+					unsigned long *strategy_parameter)
+{
+	unsigned long cacheline_size;
+	u8 byte;
+
+	pci_read_config_byte(pdev, PCI_CACHE_LINE_SIZE, &byte);
+	if (byte == 0)
+		cacheline_size = 1024;
+	else
+		cacheline_size = (int) byte * 4;
+
+	*strat = PCI_DMA_BURST_BOUNDARY;
+	*strategy_parameter = cacheline_size;
+}
+
 /* Return the index of the PCI controller for device PDEV. */
 
 extern int pci_domain_nr(struct pci_bus *bus);
--- gregkh-2.6.orig/include/asm-v850/pci.h	2005-03-01 23:37:54.000000000 -0800
+++ gregkh-2.6/include/asm-v850/pci.h	2005-06-05 22:47:10.000000000 -0700
@@ -81,6 +81,14 @@
 pci_free_consistent (struct pci_dev *pdev, size_t size, void *cpu_addr,
 		     dma_addr_t dma_addr);
 
+static inline void pci_dma_burst_advice(struct pci_dev *pdev,
+					enum pci_dma_burst_strategy *strat,
+					unsigned long *strategy_parameter)
+{
+	*strat = PCI_DMA_BURST_INFINITY;
+	*strategy_parameter = ~0UL;
+}
+
 static inline void pcibios_add_platform_entries(struct pci_dev *dev)
 {
 }
--- gregkh-2.6.orig/include/asm-x86_64/pci.h	2005-03-01 23:38:07.000000000 -0800
+++ gregkh-2.6/include/asm-x86_64/pci.h	2005-06-05 22:47:10.000000000 -0700
@@ -123,6 +123,14 @@
 	flush_write_buffers();
 }
 
+static inline void pci_dma_burst_advice(struct pci_dev *pdev,
+					enum pci_dma_burst_strategy *strat,
+					unsigned long *strategy_parameter)
+{
+	*strat = PCI_DMA_BURST_INFINITY;
+	*strategy_parameter = ~0UL;
+}
+
 #define HAVE_PCI_MMAP
 extern int pci_mmap_page_range(struct pci_dev *dev, struct vm_area_struct *vma,
 			       enum pci_mmap_state mmap_state, int write_combine);
--- gregkh-2.6.orig/include/linux/pci.h	2005-06-05 22:30:35.000000000 -0700
+++ gregkh-2.6/include/linux/pci.h	2005-06-05 22:47:10.000000000 -0700
@@ -874,6 +874,15 @@
 #define	pci_pool_alloc(pool, flags, handle) dma_pool_alloc(pool, flags, handle)
 #define	pci_pool_free(pool, vaddr, addr) dma_pool_free(pool, vaddr, addr)
 
+enum pci_dma_burst_strategy {
+	PCI_DMA_BURST_INFINITY,	/* make bursts as large as possible,
+				   strategy_parameter is N/A */
+	PCI_DMA_BURST_BOUNDARY, /* disconnect at every strategy_parameter
+				   byte boundaries */
+	PCI_DMA_BURST_MULTIPLE, /* disconnect at some multiple of
+				   strategy_parameter byte boundaries */
+};
+
 #if defined(CONFIG_ISA) || defined(CONFIG_EISA)
 extern struct pci_dev *isa_bridge;
 #endif
